/*****************************************************************************/
/*                                                                           */
/*                                 strbuf.c                                  */
/*                                                                           */
/*                       Variable sized string buffers                       */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/* (C) 2001-2012, Ullrich von Bassewitz                                      */
/*                Roemerstrasse 52                                           */
/*                D-70794 Filderstadt                                        */
/* EMail:         uz@cc65.org                                                */
/*                                                                           */
/*                                                                           */
/* This software is provided 'as-is', without any expressed or implied       */
/* warranty.  In no event will the authors be held liable for any damages    */
/* arising from the use of this software.                                    */
/*                                                                           */
/* Permission is granted to anyone to use this software for any purpose,     */
/* including commercial applications, and to alter it and redistribute it    */
/* freely, subject to the following restrictions:                            */
/*                                                                           */
/* 1. The origin of this software must not be misrepresented; you must not   */
/*    claim that you wrote the original software. If you use this software   */
/*    in a product, an acknowledgment in the product documentation would be  */
/*    appreciated but is not required.                                       */
/* 2. Altered source versions must be plainly marked as such, and must not   */
/*    be misrepresented as being the original software.                      */
/* 3. This notice may not be removed or altered from any source              */
/*    distribution.                                                          */
/*                                                                           */
/*****************************************************************************/



#include <string.h>
#include <ctype.h>

/* common */
#include "chartype.h"
#include "strbuf.h"
#include "va_copy.h"
#include "xmalloc.h"
#include "xsprintf.h"



/*****************************************************************************/
/*                                   Data                                    */
/*****************************************************************************/



/* An empty string buf */
const StrBuf EmptyStrBuf = STATIC_STRBUF_INITIALIZER;


int SB_Utf8 = 0;


/*****************************************************************************/
/*                                   Code                                    */
/*****************************************************************************/



#if !defined(HAVE_INLINE)
StrBuf* SB_Init (StrBuf* B)
/* Initialize a string buffer */
{
    *B = EmptyStrBuf;
    return B;
}
#endif



StrBuf* SB_InitFromString (StrBuf* B, const char* S)
/* Initialize a string buffer from a literal string. Beware: The buffer won't
** store a copy but a pointer to the actual string.
*/
{
    B->Allocated = 0;
    B->Len       = strlen (S);
    B->Index     = 0;
    B->Buf       = (char*) S;
    return B;
}



void SB_Done (StrBuf* B)
/* Free the data of a string buffer (but not the struct itself) */
{
    if (B->Allocated) {
        xfree (B->Buf);
    }
}



StrBuf* NewStrBuf (void)
/* Allocate, initialize and return a new StrBuf */
{
    /* Allocate a new string buffer */
    StrBuf* B = xmalloc (sizeof (StrBuf));

    /* Initialize the struct... */
    SB_Init (B);

    /* ...and return it */
    return B;
}



void FreeStrBuf (StrBuf* B)
/* Free a string buffer */
{
    /* Allow NULL pointers */
    if (B) {
        SB_Done (B);
        xfree (B);
    }
}



void SB_Realloc (StrBuf* B, unsigned NewSize)
/* Reallocate the string buffer space, make sure at least NewSize bytes are
** available.
*/
{
    /* Get the current size, use a minimum of 8 bytes */
    unsigned NewAllocated = B->Allocated;
    if (NewAllocated == 0) {
        NewAllocated = 8;
    }

    /* Round up to the next power of two */
    while (NewAllocated < NewSize) {
        NewAllocated *= 2;
    }

    /* Reallocate the buffer. Beware: The allocated size may be zero while the
    ** length is not. This means that we have a buffer that wasn't allocated
    ** on the heap.
    */
    if (B->Allocated) {
        /* Just reallocate the block */
        B->Buf   = xrealloc (B->Buf, NewAllocated);
    } else {
        /* Allocate a new block and copy */
        B->Buf   = memcpy (xmalloc (NewAllocated), B->Buf, B->Len);
    }

    /* Remember the new block size */
    B->Allocated = NewAllocated;
}



static void SB_CheapRealloc (StrBuf* B, unsigned NewSize)
/* Reallocate the string buffer space, make sure at least NewSize bytes are
** available. This function won't copy the old buffer contents over to the new
** buffer and may be used if the old contents are overwritten later.
*/
{
    /* Get the current size, use a minimum of 8 bytes */
    unsigned NewAllocated = B->Allocated;
    if (NewAllocated == 0) {
        NewAllocated = 8;
    }

    /* Round up to the next power of two */
    while (NewAllocated < NewSize) {
        NewAllocated *= 2;
    }

    /* Free the old buffer if there is one */
    if (B->Allocated) {
        xfree (B->Buf);
    }

    /* Allocate a fresh block */
    B->Buf = xmalloc (NewAllocated);

    /* Remember the new block size */
    B->Allocated = NewAllocated;
}



#if !defined(HAVE_INLINE)
char SB_At (const StrBuf* B, unsigned Index)
/* Get a character from the buffer */
{
    PRECONDITION (Index < B->Len);
    return B->Buf[Index];
}
#endif



void SB_Drop (StrBuf* B, unsigned Count)
/* Drop characters from the end of the string. */
{
    PRECONDITION (Count <= B->Len);
    B->Len -= Count;
    if (B->Index > B->Len) {
        B->Index = B->Len;
    }
}



void SB_Terminate (StrBuf* B)
/* Zero terminate the given string buffer. NOTE: The terminating zero is not
** accounted for in B->Len, if you want that, you have to use AppendChar!
*/
{
    unsigned NewLen = B->Len + 1;
    if (NewLen > B->Allocated) {
        SB_Realloc (B, NewLen);
    }
    B->Buf[B->Len] = '\0';
}



void SB_CopyBuf (StrBuf* Target, const char* Buf, unsigned Size)
/* Copy Buf to Target, discarding the old contents of Target */
{
    if (Size) {
        if (Target->Allocated < Size) {
            SB_CheapRealloc (Target, Size);
        }
        memcpy (Target->Buf, Buf, Size);
    }
    Target->Len = Size;
}



#if !defined(HAVE_INLINE)
void SB_CopyStr (StrBuf* Target, const char* S)
/* Copy S to Target, discarding the old contents of Target */
{
    SB_CopyBuf (Target, S, strlen (S));
}
#endif



#if !defined(HAVE_INLINE)
void SB_Copy (StrBuf* Target, const StrBuf* Source)
/* Copy Source to Target, discarding the old contents of Target */
{
    SB_CopyBuf (Target, Source->Buf, Source->Len);
    Target->Index = Source->Index;
}
#endif

/* Intermediate encoding for russian characters */
static int from_utf8(int C)
{
	switch (C) {
	case 0x2581: case 0x2582: case 0x2583: return 0x81;
	case 0x2584: case 0x2585: return 0x82;
	case 0x2586: case 0x2587: case 0x2588: return 0x83;
	case 0x2228: case 0x22C1: return 0x84;
	case 0x2142: case 0x2138: return 0x85;
	case 0x2200: return 0x86;
	case 0x2203: return 0x87;
	case 0x2190: case 0x21D0: case 0x21E6: case 0x21FD: return 0x88;
	case 0x256D: case 0x25DC: return 0x89;
	case 0x256E: case 0x25DD: return 0x8A;
	case 0x256F: case 0x25DE: return 0x8B;
	case 0x2570: case 0x25DF: return 0x8C;
	case 0x2BA6: case 0x2B73: return 0x8D;
	case 0x2198: case 0x21F2: return 0x8E;
	case 0x451: return 0x8F;
	case 0x250C: case 0x250D: case 0x250E: case 0x250F: case 0x2552: case 0x2553: case 0x2554: return 0x90;
	case 0x252C: case 0x252D: case 0x252E: case 0x252F: case 0x2564: case 0x2565: case 0x2566: return 0x91;
	case 0x2510: case 0x2511: case 0x2512: case 0x2513: case 0x2555: case 0x2556: case 0x2557: return 0x92;
	case 0x251C: case 0x251D: case 0x251E: case 0x251F: case 0x255E: case 0x255F: case 0x2560: return 0x93;
	case 0x253C: case 0x253D: case 0x253E: case 0x253F:
	case 0x2540: case 0x2541: case 0x2542: case 0x2543:
	case 0x2544: case 0x2545: case 0x2546: case 0x2547:
	case 0x2548: case 0x2549: case 0x254A: case 0x254B:
	case 0x256A: case 0x256B: return 0x94;
	case 0x2192: case 0x21D2: case 0x21E8: case 0x21FE: return 0x95;
	case 0x2197: case 0x21D7: return 0x96;
	case 0x2199: case 0x21D9: return 0x97;
	case 0x2196: case 0x21D6: return 0x98;
	case 0x2191: case 0x21D1: case 0x21E7: return 0x99;
	case 0x2193: case 0x21D3: case 0x21E9: return 0x9A;
	case 0x2500: case 0x2501: case 0x2550: return 0x9B;
	case 0x2524: case 0x2525: case 0x2526: case 0x2527:
	case 0x2528: case 0x2529: case 0x252A: case 0x252B:
	case 0x2561: case 0x2562: case 0x2563: return 0x9C;
	case 0x2514: case 0x2515: case 0x2516: case 0x2517:
	case 0x2558: case 0x2559: case 0x255A: return 0x9D;
	case 0x2534: case 0x2535: case 0x2536: case 0x2537:
	case 0x2538: case 0x2539: case 0x253A: case 0x253B:
	case 0x2567: case 0x2568: case 0x2569: return 0x9E;
	case 0x2518: case 0x2519: case 0x251A: case 0x251B:
	case 0x255B: case 0x255C: case 0x255D: return 0x9F;
	case 0x3B1: return 0xA1;
	case 0x3B2: return 0xA2;
	case 0x3B3: return 0xA3;
	case 0x3B5: return 0xA4;
	case 0x3BB: return 0xA5;
	case 0x3BC: return 0xA6;
	case 0x3C0: return 0xA7;
	case 0x3C1: return 0xA8;
	case 0x3C4: return 0xA9;
	case 0x3C6: return 0xAA;
	case 0x3C8: return 0xAB;
	case 0x3C9: return 0xAC;
	case 0x3A3: return 0xAD;
	case 0x394: return 0xAE;
	case 0xF7: return 0xAF;
	case 0x2070: return 0xB0;
	case 0xB9: return 0xB1;
	case 0xB2: return 0xB2;
	case 0xB3: return 0xB3;
	case 0x2074: return 0xB4;
	case 0x2075: return 0xB5;
	case 0x2076: return 0xB6;
	case 0x2077: return 0xB7;
	case 0x2078: return 0xB8;
	case 0x2079: return 0xB9;
	case 0x207B: return 0xBA;
	case 0x2261: return 0xBB;
	case 0x2264: case 0x217D: return 0xBC;
	case 0x2260: return 0xBD;
	case 0x2265: case 0x217E: return 0xBE;
	case 0xB1: return 0xBF;
	case 0xA7: return 0x60;
	case 0x22CD: return 0x7E;
	case 0x44E: return 0xC0;
	case 0x430: return 0xC1;
	case 0x431: return 0xC2;
	case 0x446: return 0xC3;
	case 0x434: return 0xC4;
	case 0x435: return 0xC5;
	case 0x444: return 0xC6;
	case 0x433: return 0xC7;
	case 0x445: return 0xC8;
	case 0x438: return 0xC9;
	case 0x439: return 0xCA;
	case 0x43A: return 0xCB;
	case 0x43B: return 0xCC;
	case 0x43C: return 0xCD;
	case 0x43D: return 0xCE;
	case 0x43E: return 0xCF;
	case 0x43F: return 0xD0;
	case 0x44F: return 0xD1;
	case 0x440: return 0xD2;
	case 0x441: return 0xD3;
	case 0x442: return 0xD4;
	case 0x443: return 0xD5;
	case 0x436: return 0xD6;
	case 0x432: return 0xD7;
	case 0x44C: return 0xD8;
	case 0x44B: return 0xD9;
	case 0x437: return 0xDA;
	case 0x448: return 0xDB;
	case 0x44D: return 0xDC;
	case 0x449: return 0xDD;
	case 0x447: return 0xDE;
	case 0x44A: return 0xDF;
	case 0xA4: return 0x24;
	case 0x42E: return 0xE0;
	case 0x410: return 0xE1;
	case 0x411: return 0xE2;
	case 0x426: return 0xE3;
	case 0x414: return 0xE4;
	case 0x415: return 0xE5;
	case 0x424: return 0xE6;
	case 0x413: return 0xE7;
	case 0x425: return 0xE8;
	case 0x418: return 0xE9;
	case 0x419: return 0xEA;
	case 0x41A: return 0xEB;
	case 0x41B: return 0xEC;
	case 0x41C: return 0xED;
	case 0x41D: return 0xEE;
	case 0x41E: return 0xEF;
	case 0x41F: return 0xF0;
	case 0x42F: return 0xF1;
	case 0x420: return 0xF2;
	case 0x421: return 0xF3;
	case 0x422: return 0xF4;
	case 0x423: return 0xF5;
	case 0x416: return 0xF6;
	case 0x412: return 0xF7;
	case 0x42C: return 0xF8;
	case 0x42B: return 0xF9;
	case 0x417: return 0xFA;
	case 0x428: return 0xFB;
	case 0x42D: return 0xFC;
	case 0x429: return 0xFD;
	case 0x427: return 0xFE;
	case 0x42A: return 0xFF;
	default: return C;
	}
}

static void RealAppendChar (StrBuf* B, int C)
/* Append a character to a string buffer */
{
	unsigned NewLen = B->Len + 1;
	if (NewLen > B->Allocated) {
		SB_Realloc (B, NewLen);
	}
	B->Buf[B->Len] = (char) C;
	B->Len = NewLen;
}

static void Utf8AppendChar(struct StrBuf* B, int C)
{
	if (B->Rem) {
		B->Ch <<= 6;
		B->Ch |= C & 0x3F;
		--B->Rem;
		if (!B->Rem) RealAppendChar(B, from_utf8(B->Ch));
	} else {
		if (C & 0x80) {
			if (C & 0x40) {
				if (C & 0x20) {
					if (C & 0x10) {
						if (C & 0x08) { // bad sequence
//							fprintf(stderr, "(1) bad utf-8 char %02X\n", C);
						} else { // 4-byte sequence
							B->Ch = C & 0x07;
							B->Rem = 3;
						}
					} else { // 3-byte sequence
						B->Ch = C & 0x0F;
						B->Rem = 2;
					}
				} else {
					B->Ch = C & 0x01F;
					B->Rem = 1;
				}
			} else {
//				fprintf(stderr, "(2) bad utf-8 char %02X\n", C);
			}
		} else {
			RealAppendChar(B, from_utf8(C));
		}
	}
}


void SB_AppendChar (StrBuf* B, int C)
/* Append a character to a string buffer */
{
	RealAppendChar(B, C);
}

void SB_AppendSourceChar (StrBuf* B, int C)
/* Append a character from a source file to a string buffer */
{
	if (SB_Utf8 && C != -1) Utf8AppendChar(B, C);
	else RealAppendChar(B, C);
}


void SB_AppendBuf (StrBuf* B, const char* S, unsigned Size)
/* Append a character buffer to the end of the string buffer */
{
    unsigned NewLen = B->Len + Size;
    if (NewLen > B->Allocated) {
        SB_Realloc (B, NewLen);
    }
    memcpy (B->Buf + B->Len, S, Size);
    B->Len = NewLen;
}



#if !defined(HAVE_INLINE)
void SB_AppendStr (StrBuf* B, const char* S)
/* Append a string to the end of the string buffer */
{
    SB_AppendBuf (B, S, strlen (S));
}
#endif



#if !defined(HAVE_INLINE)
void SB_Append (StrBuf* Target, const StrBuf* Source)
/* Append the contents of Source to Target */
{
    SB_AppendBuf (Target, Source->Buf, Source->Len);
}
#endif



#if !defined(HAVE_INLINE)
void SB_Cut (StrBuf* B, unsigned Len)
/* Cut the contents of B at the given length. If the current length of the
** buffer is smaller than Len, nothing will happen.
*/
{
    if (Len < B->Len) {
        B->Len = Len;
    }
}
#endif



void SB_Slice (StrBuf* Target, const StrBuf* Source, unsigned Start, unsigned Len)
/* Copy a slice from Source into Target. The current contents of Target are
** destroyed. If Start is greater than the length of Source, or if Len
** characters aren't available, the result will be a buffer with less than Len
** bytes.
*/
{
    /* Calculate the length of the resulting buffer */
    if (Start >= Source->Len) {
        /* Target will be empty */
        SB_Clear (Target);
        return;
    }
    if (Start + Len > Source->Len) {
        Len = Source->Len - Start;
    }

    /* Make sure we have enough room in the target string buffer */
    if (Len > Target->Allocated) {
        SB_Realloc (Target, Len);
    }

    /* Copy the slice */
    memcpy (Target->Buf, Source->Buf + Start, Len);
    Target->Len = Len;
}



void SB_Move (StrBuf* Target, StrBuf* Source)
/* Move the complete contents of Source to target. This will delete the old
** contents of Target, and Source will be empty after the call.
*/
{
    /* Free the target string */
    if (Target->Allocated) {
        xfree (Target->Buf);
    }

    /* Move all data from Source to Target */
    *Target = *Source;

    /* Clear Source */
    SB_Init (Source);
}



void SB_ToLower (StrBuf* S)
/* Convert all characters in S to lower case */
{
    unsigned I;
    char* B = S->Buf;
    for (I = 0; I < S->Len; ++I, ++B) {
        if (IsUpper (*B)) {
            *B = tolower (*B);
        }
    }
}



void SB_ToUpper (StrBuf* S)
/* Convert all characters in S to upper case */
{
    unsigned I;
    char* B = S->Buf;
    for (I = 0; I < S->Len; ++I, ++B) {
        if (IsLower (*B)) {
            *B = toupper (*B);
        }
    }
}



int SB_Compare (const StrBuf* S1, const StrBuf* S2)
/* Do a lexical compare of S1 and S2. See strcmp for result codes. */
{
    int Result;
    if (S1->Len < S2->Len) {
        Result = memcmp (S1->Buf, S2->Buf, S1->Len);
        if (Result == 0) {
            /* S1 considered lesser because it's shorter */
            Result = -1;
        }
    } else if (S1->Len > S2->Len) {
        Result = memcmp (S1->Buf, S2->Buf, S2->Len);
        if (Result == 0) {
            /* S2 considered lesser because it's shorter */
            Result = 1;
        }
    } else {
        Result = memcmp (S1->Buf, S2->Buf, S1->Len);
    }
    return Result;
}



int SB_CompareStr (const StrBuf* S1, const char* S2)
/* Do a lexical compare of S1 and S2. See strcmp for result codes. */
{
    int Result;
    unsigned S2Len = strlen (S2);
    if (S1->Len < S2Len) {
        Result = memcmp (S1->Buf, S2, S1->Len);
        if (Result == 0) {
            /* S1 considered lesser because it's shorter */
            Result = -1;
        }
    } else if (S1->Len > S2Len) {
        Result = memcmp (S1->Buf, S2, S2Len);
        if (Result == 0) {
            /* S2 considered lesser because it's shorter */
            Result = 1;
        }
    } else {
        Result = memcmp (S1->Buf, S2, S1->Len);
    }
    return Result;
}



void SB_VPrintf (StrBuf* S, const char* Format, va_list ap)
/* printf function with S as target. The function is safe, which means that
** the current contents of S are discarded, and are allocated again with
** a matching size for the output. The function will call FAIL when problems
** are detected (anything that let xsnprintf return -1).
*/
{
    va_list tmp;
    int SizeNeeded;

    /* Since we must determine the space needed anyway, we will try with
    ** the currently allocated memory. If the call succeeds, we've saved
    ** an allocation. If not, we have to reallocate and try again.
    */
    va_copy (tmp, ap);
    SizeNeeded = xvsnprintf (S->Buf, S->Allocated, Format, tmp);
    va_end (tmp);

    /* Check the result, the xvsnprintf function should not fail */
    CHECK (SizeNeeded >= 0);

    /* Check if we must reallocate */
    if ((unsigned) SizeNeeded >= S->Allocated) {
        /* Must retry. Use CheapRealloc to avoid copying */
        SB_CheapRealloc (S, SizeNeeded + 1);    /* Account for '\0' */
        (void) xvsnprintf (S->Buf, S->Allocated, Format, ap);
    }

    /* Update string buffer variables */
    S->Len = SizeNeeded;
    S->Index = 0;
}



void SB_Printf (StrBuf* S, const char* Format, ...)
/* vprintf function with S as target. The function is safe, which means that
** the current contents of S are discarded, and are allocated again with
** a matching size for the output. The function will call FAIL when problems
** are detected (anything that let xsnprintf return -1).
*/
{
    va_list ap;
    va_start (ap, Format);
    SB_VPrintf (S, Format, ap);
    va_end (ap);
}
